__version__ = '0.12.2'

from warnings import warn
# load configuration file using utilix
try:
    from utilix import uconfig
# if no utilix config, get a RuntimeError
# we don't want this to break straxen, but it does print a warning statement
except (FileNotFoundError, RuntimeError) as e:
    warn('Warning, utilix config file not loaded properly. copy '
         '/project2/lgrandi/xenonnt/.xenon_config to your HOME directory!',
         UserWarning)
    uconfig = None

from .common import *
from .itp_map import *
from .rundb import *
from .online_monitor import *
from .matplotlib_utils import *
from .mini_analysis import *
from .misc import *
from .scada import *

from .get_corrections import *
from .hitfinder_thresholds import *
from .corrections_services import *

from . import plugins
from .plugins import *

from . import analyses

# Do not make all contexts directly available under straxen.
# Otherwise we have straxen.demo() etc.
from . import contexts



import strax
import numpy as np
import numexpr

export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option(
        'area_vs_width_nbins',
        type=int, default=60,
        help='Number of bins for area vs width histogram for online monitor. '
             'NB: this is a 2D histogram'),
    strax.Option(
        'area_vs_width_bounds',
        type=tuple, default=((0, 5), (0, 5)),
        help='Boundaries of log-log histogram of area vs width'),
    strax.Option(
        'area_vs_width_min_gap',
        type=int, default=20,
        help='Minimal gap between consecutive peaks to be considered for the '
             '"area_vs_width_hist_clean" To turn off this cut, set to 0.'),
    strax.Option(
        'area_vs_width_cut_string',
        type=str, default='',
        help='Selection (like selection_str) applied to data for '
             '"area_vs_width_hist_clean", cuts should be separated using "&"'
             'For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)'
             'Default is no selection (other than "area_vs_width_min_gap")'),
    strax.Option(
        'lone_hits_area_bounds',
        type=tuple, default=(0, 1500),
        help='Boundaries area histogram of lone hits [ADC]'),
    strax.Option(
        'online_monitor_nbins',
        type=int, default=100,
        help='Number of bins of histogram of online monitor. Will be used '
             'for: '
             'lone_hits_area-histogram, '
             'area_fraction_top-histogram, '
             'near_s1_hists, '),
    strax.Option(
        'near_s1_hists_cut_string',
        type=str,
        default='(n_channels > 20) & (n_channels < 400) & (area < 1000) & '
                '(area > 5) & (rise_time < 100) & (type == 1)',
        help='Selection (like selection_str) applied to data for '
             '"near_s1_hists", cuts should be separated using "&"'
             'For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)'
             'Default is no selection (other than "area_vs_width_min_gap")'),
    strax.Option(
        'near_s1_hists_bounds',
        type=tuple,
        default=(0, 1000),
        help='Bounds for the near s1-peaks in PE'),
    strax.Option(
        'near_s1_max_time_diff',
        type=int, default=2_000,
        help='Max gap between two peaks for the near-s1 area histogram [ns]'),
    strax.Option(
        'n_tpc_pmts', type=int,
        help='Number of TPC PMTs'),
)
class OnlinePeakMonitor(strax.Plugin):
    """
    Plugin to write data to the online-monitor. Data that is written by
    this plugin should be small such as to not overload the runs-
    database.

    This plugin takes 'peaks_basics' and 'lone_hits'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in hte online monitor.

    Produces 'online_peak_monitor' with info on the lone-hits and peaks
    """
    depends_on = ('peak_basics', 'lone_hits')
    provides = 'online_peak_monitor'
    __version__ = '0.0.4'
    # TODO make new datakind:
    # data_kind = 'online_monitor'
    rechunk_on_save = False

    def infer_dtype(self):
        n_bins_area_width = self.config['area_vs_width_nbins']
        bounds_area_width = self.config['area_vs_width_bounds']

        n_bins = self.config['online_monitor_nbins']

        n_tpc_pmts = self.config['n_tpc_pmts']
        dtype = [
            (('Start time of the chunk', 'time'),
             np.int64),
            (('End time of the chunk', 'endtime'),
             np.int64),
            (('Area vs width histogram (log-log)', 'area_vs_width_hist'),
             (np.int64, (n_bins_area_width, n_bins_area_width))),
            (('Area vs width edges (log-space)', 'area_vs_width_bounds'),
             (np.float64, np.shape(bounds_area_width))),
            (('Area vs width histogram with cuts (log-log)', 'area_vs_width_hist_clean'),
             (np.int64, (n_bins_area_width, n_bins_area_width))),
            (('Lone hits areas histogram [ADC-counts]', 'lone_hits_area_hist'),
             (np.int64, n_bins)),
            (('Lone hits areas bounds [ADC-counts]', 'lone_hits_area_bounds'),
             (np.float64, 2)),
            (('Lone hits per channel', 'lone_hits_per_channel'),
             (np.int64, n_tpc_pmts)),
            (('AFT histogram', 'aft_hist'),
             (np.int64, n_bins)),
            (('AFT bounds', 'aft_bounds'),
             (np.float64, 2)),
            (('Number of contributing channels histogram', 'n_channel_hist'),
             (np.int64, n_tpc_pmts)),
            (('Number of contributing channels histogram bounds', 'n_channel_bounds'),
             (np.float64, 2)),
            (('Near S1 peaks area hist', 'near_s1_area_hist'),
             (np.int64, n_bins)),
            (('Near S1 peaks area hist bounds', 'near_s1_area_bounds'),
             (np.float64, 2)),
        ]
        return dtype

    def compute(self, peaks, lone_hits, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end
        n_pmt = self.config['n_tpc_pmts']
        n_bins = self.config['online_monitor_nbins']

        # Bounds for histograms
        res['area_vs_width_bounds'] = self.config['area_vs_width_bounds']
        res['lone_hits_area_bounds'] = self.config['lone_hits_area_bounds']

        # -- Peak vs area 2D histogram --
        # Always cut out unphysical peaks
        sel = (peaks['area'] > 0) & (peaks['range_50p_area'] > 0)
        res['area_vs_width_hist'] = self.area_width_hist(peaks[sel])

        # Experimental example of how to apply cuts here.
        # Let's make a cut on the time between two peaks, if too short,
        # ignore the peak.
        timedelta = peaks[1:]['time'] - strax.endtime(peaks)[:-1]
        timesel = timedelta > self.config['area_vs_width_min_gap']
        # Last peak always has no tails
        timesel = np.concatenate((timesel, [True]))
        sel &= timesel

        # Also apply the area_vs_width_cut_string like a selection_str
        sel_str = self.config['area_vs_width_cut_string']
        sel = self._config_as_selection_str(sel_str, peaks, pre_sel=sel)
        res['area_vs_width_hist_clean'] = self.area_width_hist(peaks[sel])
        # make a new selection don't re-use
        del sel

        # -- Lone hit properties --
        # Make histogram of ADC counts
        lone_hit_areas, _ = np.histogram(lone_hits['area'],
                                         bins=n_bins,
                                         range=self.config['lone_hits_area_bounds'])

        lone_hit_channel_count, _ = np.histogram(lone_hits['channel'],
                                                 bins=n_pmt,
                                                 range=[0, n_pmt])
        # Count number of lone-hits per PMT
        res['lone_hits_area_hist'] = lone_hit_areas
        res['lone_hits_per_channel'] = lone_hit_channel_count

        # -- AFT histogram --
        aft_b = [0, 1]
        aft_hist, _ = np.histogram(peaks['area_fraction_top'], bins=n_bins, range=aft_b)
        res['aft_hist'] = aft_hist
        res['aft_bounds'] = aft_b

        # -- Number of contributing channels channels --
        n_cont_b = [0, n_pmt]
        n_cont_hist, _ = np.histogram(peaks['n_channels'], bins=n_pmt, range=n_cont_b)
        res['n_channel_hist'] = n_cont_hist
        res['n_channel_bounds'] = n_cont_b

        # -- Experimental selection --
        # We first apply a basic selection on the peaks to e.g. get S1s
        mask = self._config_as_selection_str(self.config['near_s1_hists_cut_string'], peaks)
        peaks_sel = peaks[mask]
        # We select peaks where the peak before or the peak after it is within
        # near_s1_max_time_diff ns. TODO: Do we want another hist for this?
        time_diff = peaks_sel[1:]['time'] - strax.endtime(peaks_sel)[:-1]
        is_close = time_diff < self.config['near_s1_max_time_diff']
        time_mask = np.zeros(len(peaks_sel), dtype=np.bool_)
        # Either the previous or the next peak can be close, take both into account
        time_mask[:-1] = is_close
        time_mask[1:] = time_mask[1:] | is_close

        # Make the area hist
        near_s1_bound = self.config['near_s1_hists_bounds']
        near_s1_hist, _ = np.histogram(peaks_sel['area'][time_mask], bins=n_bins, range=near_s1_bound)
        res['near_s1_area_hist'] = near_s1_hist
        res['near_s1_area_bounds'] = near_s1_bound

        # Cleanup
        # del hist, clean_hist, lone_hit_areas, lone_hit_channel_count
        return res

    # TODO
    #  somehow prevent overlap with strax.context.apply_selection
    @staticmethod
    def _config_as_selection_str(selection_string, data, pre_sel=None):
        """Get mask for data base on the selection string"""
        if pre_sel is None:
            pre_sel = np.ones(len(data), dtype=np.bool_)

        if selection_string != '':
            if isinstance(selection_string, (list, tuple)):
                selection_string = ' & '.join(f'({x})' for x in selection_string)

            mask = numexpr.evaluate(selection_string, local_dict={
                fn: data[fn]
                for fn in data.dtype.names})
            pre_sel &= mask
        return pre_sel

    def area_width_hist(self, data):
        """Make area vs width 2D-hist"""
        hist, _, _ = np.histogram2d(
            np.log10(data['area']),
            np.log10(data['range_50p_area']),
            range=self.config['area_vs_width_bounds'],
            bins=self.config['area_vs_width_nbins'])
        return hist.T


class OnlineMonitor(strax.LoopPlugin):
    """
    Loop over the online-monitor chunks, get the veto intervals that are within
    each of these chunks. Compute the live-time within each of the chunks.
    """
    depends_on = ('online_peak_monitor', 'veto_intervals')
    provides = 'online_monitor'
    __version__ = '0.0.4'
    rechunk_on_save = False

    def infer_dtype(self):
        dtype = strax.unpack_dtype(self.deps['online_peak_monitor'].dtype_for('online_peak_monitor'))
        dtype += [(('Live time', 'live_time'),
                   np.float64),]
        return dtype

    def compute_loop(self, peaks, veto_intervals):
        res = {}
        for d in peaks.dtype.names:
            res[d] = peaks[d]
        dt = strax.endtime(peaks) - peaks['time']
        assert not np.iterable(dt) or len(dt) == 1
        if dt > 0:
            res['live_time'] = 1 - np.sum(veto_intervals['veto_interval'])/dt
        else:
            res['live_time'] = 1
        return res



"""Return corrections from corrections DB
"""
import pytz
import numpy as np
from functools import lru_cache
import configparser
import strax
import straxen
from straxen import uconfig

export, __all__ = strax.exporter()


@export
class CorrectionsManagementServices():
    """
    A class that returns corrections
    Corrections are set of parameters to be applied in the analysis
    stage to remove detector effects. Information on the strax implementation
    can be found at https://github.com/AxFoundation/strax/blob/master/strax/corrections.py
    """
    def __init__(self, username=None, password=None, mongo_url=None, is_nt=True):
        """
        :param username: corrections DB username
            read the .xenon_config for the users "pymongo_user" has
            readonly permissions to the corrections DB
            the "CMT admin user" has r/w permission to corrections DB
            and read permission to runsDB
        :param password: DB password
        :param is_nt: bool if True we are looking at nT if False we are looking at 1T
        """
        # TODO avoid duplicated code with the RunDB.py?
        # Basic setup
        if username is not None:
            self.username = username
        else:
            self.username = uconfig.get('RunDB', 'pymongo_user')
        if password is not None:
            self.password = password
        else:
            self.password = uconfig.get('RunDB', 'pymongo_password')

        if mongo_url is None:
            mongo_url = uconfig.get('RunDB', 'pymongo_url')

        # Setup the interface
        self.interface = strax.CorrectionsInterface(
            host=f'mongodb://{mongo_url}',
            username=self.username,
            password=self.password,
            database_name='corrections')
        # Use the same client as the CorrectionsInterface
        client = self.interface.client

        self.is_nt = is_nt
        if self.is_nt:
            self.collection = client['xenonnt']['runs']
        else:
            self.collection = client['run']['runs_new']

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return str(f'{"XENONnT " if self.is_nt else "XENON1T"}'
                   f'-Corrections_Management_Services')

    def get_corrections_config(self, run_id, correction=None, config_model=None):
        """
        Get context configuration for a given correction
        :param run_id: run id from runDB
        :param correction: correction's name (str type)
        :param config_model: configuration model (tuple type)
        :return: correction value(s)
        """

        if not isinstance(config_model, (tuple, list)) or len(config_model) != 2:
            raise ValueError(f'config_model {config_model} must be a tuple')
        model_type, global_version = config_model

        if correction == 'pmt_gains':
            return self.get_pmt_gains(run_id, model_type, global_version)
        elif correction == 'elife':
            return self.get_elife(run_id, model_type, global_version)
        else:
            raise ValueError(f'{correction} not found')

    # TODO add option to extract 'when'. Also, the start time might not be the best
    #  entry for e.g. for super runs
    # cache results, this would help when looking at the same gains
    @lru_cache(maxsize=None)
    def _get_correction(self, run_id, correction, global_version):
        """
        Smart logic to get correction from DB
        :param run_id: run id from runDB
        :param correction: correction's name, key word (str type)
        :param global_version: global version (str type)
        :return: correction value(s)
        """
        when = self.get_start_time(run_id)
        df_global = self.interface.read('global' if self.is_nt else 'global_xenon1t')

        try:
            values = []
            for it_correction, version in df_global.iloc[-1][global_version].items():
                if correction in it_correction:
                    df = self.interface.read(it_correction)
                    if global_version == 'ONLINE':
                        # We don't want to have different versions based
                        # on when something was processed therefore
                        # don't interpolate but forward fill.
                        df = self.interface.interpolate(df, when, how='fill')
                    else:
                        df = self.interface.interpolate(df, when)
                    values.append(df.loc[df.index == when, version].values[0])
            corrections = np.asarray(values)
        except KeyError:
            raise ValueError(f'Global version {global_version} not found for correction {correction}')

        # for single value corrections, e.g. elife correction
        if len(corrections) == 1:
            return float(corrections)
        else:
            return corrections

    def get_elife(self, run_id, model_type, global_version):
        """
        Smart logic to return electron lifetime correction
        :param run_id: run id from runDB
        :param model_type: choose either elife_model or elife_constant
        :param global_version: global version, or float (if model_type == elife_constant) 
        :return: electron lifetime correction value
        """
        if model_type == 'elife_model':
            return self._get_correction(run_id, 'elife', global_version)

        elif model_type == 'elife_constant':
            # This is nothing more than just returning the value we put in
            if not isinstance(global_version, float):
                raise ValueError(f'User specify a model type {model_type} '
                                 f'and should provide a float. Got: '
                                 f'{type(global_version)}')
            return float(global_version)

        else:
            raise ValueError(f'model type {model_type} not implemented for electron lifetime')

    # TODO create a propper dict for 'to_pe_constant' and 'global_version' as
    #  the 'global_version' is not a version but an array/float for
    #  model_type = 'to_pe_constant'
    def get_pmt_gains(self, run_id, model_type, global_version, gain_dtype = np.float32):
        """
        Smart logic to return pmt gains to PE values.
        :param run_id: run id from runDB
        :param model_type: Choose either to_pe_model or to_pe_constant
        :param global_version: global version or a constant value or an array (if
        model_type == to_pe_constant)
        :param gain_dtype: dtype of the gains to be returned as array
        :return: array of pmt gains to PE values
        """
        if model_type == 'to_pe_model':
            to_pe = self._get_correction(run_id, 'pmt', global_version)
            # be cautious with very early runs, check that not all are None
            if np.isnan(to_pe).all():
                raise ValueError(
                        f'to_pe(PMT gains) values are NaN, no data available'
                        f' for {run_id} in the gain model with version '
                        f'{global_version}, please set constant values for '
                        f'{run_id}')

        elif model_type == 'to_pe_constant':
            n_tpc_pmts = straxen.n_tpc_pmts
            if not self.is_nt:
                # TODO can we prevent these kind of hard codes using the context?
                n_tpc_pmts = straxen.contexts.x1t_common_config['n_tpc_pmts']

            if not isinstance(global_version, (float, np.ndarray, int)):
                raise ValueError(f'User must specify a model type {model_type} '
                                 f'and provide a float/array to be used. Got: '
                                 f'{type(global_version)}')

            # Generate an array of values and multiply by the 'global_version'
            to_pe = np.ones(n_tpc_pmts, dtype=gain_dtype) * global_version
            if len(to_pe) != n_tpc_pmts:
                raise ValueError(f'to_pe length does not match {n_tpc_pmts}. '
                                 f'Check that {global_version} is either of '
                                 f'length {n_tpc_pmts} or a float')

        else:
            raise ValueError(f'{model_type} not implemented for to_pe values')

        # Double check the dtype of the gains
        to_pe = np.array(to_pe, dtype=gain_dtype)

        # Double check that all the gains are found, None is not allowed
        # since strax processing does not handle this well. If a PMT is
        # off it's gain should be 0.
        if np.any(np.isnan(to_pe)):
            pmts_affected = np.argwhere(np.isnan(to_pe))[:, 0]
            raise GainsNotFoundError(
                f'Gains returned by CMT are None for PMT_i = {pmts_affected}. '
                f'Cannot proceed with processing. Report to CMT-maintainers.')
        return to_pe

    def get_lce(self, run_id, s, position, global_version='v1'):
        """
        Smart logic to return light collection eff map values.
        :param run_id: run id from runDB
        :param s: S1 map or S2 map
        :param global_version:
        :param position: event position
        """
        raise NotImplementedError

    def get_fdc(self, run_id, position, global_version='v1'):
        """
        Smart logic to return field distortion map values.
        :param run_id: run id from runDB
        :param position: event position
        :param global_version: global version (str type)
        """
        raise NotImplementedError

    # TODO change to st.estimate_start_time
    def get_start_time(self, run_id):
        """
        Smart logic to return start time from runsDB
        :param run_id: run id from runDB
        :return: run start time
        """

        if self.is_nt:
            # xenonnt use int
            run_id = int(run_id)

        rundoc = self.collection.find_one(
                {'number' if self.is_nt else 'name': run_id},
                {'start': 1})
        if rundoc is None:
            raise ValueError(f'run_id = {run_id} not found')
        time = rundoc['start']
        return time.replace(tzinfo=pytz.utc)


class GainsNotFoundError(Exception):
    """Fatal error if a None value is returned by the corrections"""
    pass



import urllib
import requests
import pandas as pd
import numba
import numpy as np
import warnings

import strax
import straxen

from straxen import uconfig

export, __all__ = strax.exporter()


@export
class SCADAInterface:

    def __init__(self, context=None):
        """
        Interface to excess the XENONnT slow control data via python.

        :param context: Context you are using e.g. st. This is needed
            if you would like to query data via run_ids.
        """
        try:
            self.SCData_URL = uconfig.get('scada', 'scdata_url')
            self.SCLastValue_URL = uconfig.get('scada', 'sclastvalue_url')
            self.SCADA_SECRETS = dict(QueryType=uconfig.get('scada', 'querytype'),
                                      username=uconfig.get('scada', 'username'),
                                      api_key=uconfig.get('scada', 'api_key')
                                      )
        except ValueError:
            raise ValueError(f'Cannot load SCADA information, from your xenon'
                             ' config. SCADAInterface cannot be used.')
        self.context = context

    def get_scada_values(self,
                         parameters,
                         start=None,
                         end=None,
                         run_id=None,
                         time_selection_kwargs={'full_range': True},
                         interpolation=False,
                         filling_kwargs={},
                         down_sampling=False,
                         value_every_seconds=1):
        """
        Function which returns XENONnT slow control values for a given
        set of parameters and time range.

        The time range can be either defined by a start and end time or
        via the run_id, target and context.

        :param parameters: dictionary containing the names of the
            requested scada-parameters. The keys are used as identifier
            of the parameters in the returned pandas.DataFrame.
        :param start: int representing the start time of the interval
            in ns unix time.
        :param end: same as start but as end.
        :param run_id: Id of the run. Can also be specified as a list or
            tuple of run ids. In this case we will return the time
            range lasting between the start of the first and endtime
            of the second run.
        :param time_selection_kwargs: Keyword arguments taken by
            st.to_absolute_time_range(). Default: full_range=True.
        :param interpolation: Boolean which decided to either forward
            fill empty values or to interpolate between existing ones.
        :param filling_kwargs: Kwargs applied to pandas .ffill() or
            .interpolate().
        :param down_sampling: Boolean which indicates whether to
            donw_sample result or to apply moving average. Moving average
            is deactivated in case of interpolated data.
        :param value_every_seconds: Defines with which time difference
            values should be returned. Must be an integer!
            Default: one value per 1 seconds.
        :return: pandas.DataFrame containing the data of the specified
            parameters.
        """
        if not isinstance(parameters, dict):
            mes = 'The argument "parameters" has to be specified as a dict.'
            raise ValueError(mes)

        if np.all((run_id, self.context)):
            # User specified a valid context and run_id, so get the start
            # and end time for our query:
            if isinstance(run_id, (list, tuple)):
                run_id = np.sort(run_id)  # Do not trust the user's
                start, _ = self.context.to_absolute_time_range(run_id[0], **time_selection_kwargs)
                _, end = self.context.to_absolute_time_range(run_id[-1], **time_selection_kwargs)
            else:
                start, end = self.context.to_absolute_time_range(run_id, **time_selection_kwargs)
        elif np.any(run_id):
            mes = ('You are trying to query slow control data via run_ids' 
                  ' but you have not specified the context you are '
                   'working with. Please set the context either via '
                   '.st = YOURCONTEXT, or when initializing the '
                   'interface.')
            raise ValueError(mes)

        if not np.all((start, end)):
            # User has not specified any vaild start and end time
            mes = ('You have to specify either a run_id and context.'
                   ' E.g. call get_scada_values(parameters, run_id=run)'
                   ' or you have to specifiy a valid start and end time '
                   'in utc unix time ns.')
            raise ValueError(mes)

        now = np.datetime64('now')
        if (end // 10**9) > now.astype(np.int64):
            mes = ('You are asking for an endtime which is in the future,'
                   ' I may be written by a physicist, but I am neither self-'
                   'aware nor can I predict the future like they can. You '
                   f'asked for the endtime: {end // 10**9} but current utc '
                   f'time is {now.astype(np.int64)}. I will return for the values for the '
                   'corresponding times as nans instead.')
            warnings.warn(mes)

        # Now loop over specified parameters and get the values for those.
        for ind, (k, p) in enumerate(parameters.items()):
            print(f'Start to query {k}: {p}')
            temp_df = self._query_single_parameter(start, end,
                                                   k, p,
                                                   value_every_seconds=value_every_seconds,
                                                   interpolation=interpolation,
                                                   filling_kwargs=filling_kwargs,
                                                   down_sampling=down_sampling
                                                   )

            if ind:
                m = np.all(df.loc[:, 'time'] == temp_df.loc[:, 'time'])
                mes = ('This is odd somehow the time stamps for the query of'
                       f' {p} does not match the other time stamps.')
                assert m, mes
                df = pd.concat((df, temp_df[k]), axis=1)
            else:
                df = temp_df

        # Adding timezone information and rename index:
        df.set_index('time', inplace=True)
        df = df.tz_localize(tz='UTC')
        df.index.rename('time UTC', inplace=True)

        if (end // 10**9) > now.astype(np.int64):
            df.loc[now:, :] = np.nan

        return df

    def _query_single_parameter(self,
                                start,
                                end,
                                parameter_key,
                                parameter_name,
                                interpolation,
                                filling_kwargs,
                                down_sampling,
                                value_every_seconds=1):
        """
        Function to query the values of a single parameter from SCData.

        :param start: Start time in ns unix time
        :param end: End time in ns unix time
        :param parameter_key: Key to identify queryed parameter in the
            DataFrame
        :param parameter_name: Parameter name in Scada/historian database.
        :param value_every_seconds: Defines with which time difference
            values should be returned. Must be an integer!
            Default: one value per 1 seconds.

        :returns: DataFrame with a time and parameter_key column.
        """
        if value_every_seconds < 1:
            mes = ("Scada takes only values every second. Cannot ask for a"
                   " higher sampling rate than one value per second. However"
                   f" you asked for one value every {value_every_seconds} seconds.")
            raise ValueError(mes)
        if not isinstance(value_every_seconds, int):
            raise ValueError('"value_every_seconds" must be an int!')

        # First we have to create an array where we can fill values with
        # the sampling frequency of scada:
        # TODO: Add a check in case user queries to many values. If yes read
        #  the data in chunks. How much are too many?
        seconds = np.arange(start, end + 1, 10**9)  # +1 to make sure endtime is included
        df = pd.DataFrame()
        df.loc[:, 'time'] = seconds
        df['time'] = df['time'].astype('<M8[ns]')
        df.set_index('time', inplace=True)

        # Check if first value is in requested range:
        query = self.SCADA_SECRETS.copy()
        query['name'] = parameter_name
        query['EndDateUnix'] = (start // 10**9) + 1  # +1 since it is end before exclusive
        query = urllib.parse.urlencode(query)
        values = requests.get(self.SCLastValue_URL + query)

        try:
            temp_df = pd.read_json(values.text)
        except ValueError:
            mes = values.text  # returns a dictionary as a string
            mes = eval(mes)
            raise ValueError(f'SCADA raised the following error "{mes["message"]}" '
                             f'when looking for your parameter "{parameter_name}"')

        # Store value as first value in our df
        df.loc[df.index.values[0], parameter_key] = temp_df['value'][0]

        # Query values between start+1 and end time:
        query = self.SCADA_SECRETS.copy()
        query["StartDateUnix"] = (start // 10**9) + 1
        query["EndDateUnix"] = (end // 10**9)
        query['name'] = parameter_name
        query = urllib.parse.urlencode(query)
        values = requests.get(self.SCData_URL + query)

        try:
            # Here we cannot do any better since the Error message returned
            # by the scada api is always the same...
            temp_df = pd.read_json(values.text)
            df.loc[temp_df['timestampseconds'], parameter_key] = temp_df.loc[:, 'value'].values
        except ValueError:
            pass

        # Let user decided whether to ffill or interpolate:
        if interpolation:
            df.interpolate(**filling_kwargs, inplace=True)
        else:
            # Now fill values in between like Scada would do:
            df.ffill(**filling_kwargs, inplace=True)
        df.reset_index(inplace=True)

        # Step 4. Down-sample data if asked for:
        if value_every_seconds > 1:
            if interpolation and not down_sampling:
                warnings.warn('Cannot use interpolation and running average at the same time.'
                              ' Deactivated the running average, switch to down_sampling instead.')
                down_sampling = True

            if down_sampling:
                df = df[::value_every_seconds]
            else:
                nt, nv = _average_scada(df['time'].astype(np.int64).values,
                                        df[parameter_key].values,
                                        value_every_seconds)
                df = pd.DataFrame()
                df['time'] = nt.astype('<M8[ns]')
                df[parameter_key] = nv

        return df

    def find_scada_parameter(self):
        # TODO: Add function which returns SCADA sensor names by short Name
        raise NotImplementedError('Feature not implemented yet.')


@export
def convert_time_zone(df, tz):
    """
    Function which converts the current time zone of a given
    pd.DataFrame into another timezone.

    :param df: pandas.DataFrame containing the Data. Index must be a
        datetime object with time zone information.
    :param tz: str representing the timezone the index should be
        converted to. See the notes for more information.
    :return: pandas.DataFrame with converted time index.

    Notes:
        1. ) The input pandas.DataFrame must be indexed via datetime
        objects which are timezone aware.

        2.)  You can find a complete list of available timezones via:
        ```
        import pytz
        pytz.all_timezones
        ```
        You can also specify 'strax' as timezone which will convert the
        time index into a 'strax time' equivalent.
        The default timezone of strax is UTC.
    """
    if tz == 'strax':
        df = df.tz_convert(tz='UTC')
        df.index = df.index.astype(np.int64)
        df.index.rename(f'time strax', inplace=True)
    else:
        df = df.tz_convert(tz=tz)
        df.index.rename(f'time {tz}', inplace=True)
    return df


@numba.njit
def _average_scada(times, values, nvalues):
    """
    Function which down samples scada values.

    :param times: Unix times of the data points.
    :param values: Corresponding sensor value
    :param nvalues: Number of samples we average over.
    :return: new time values and
    """
    if len(times) % nvalues:
        nsamples = (len(times) // nvalues) - 1
    else:
        nsamples = (len(times) // nvalues)
    res = np.zeros(nsamples, dtype=np.float32)
    new_times = np.zeros(nsamples, dtype=np.int64)
    for ind in range(nsamples):
        res[ind] = np.mean(values[ind * nvalues:(ind + 1) * nvalues])
        new_times[ind] = np.mean(times[ind * nvalues:(ind + 1) * nvalues])

    return new_times, res
